from arcgis.gis import GIS, Item
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict
from datetime import datetime
from collections import OrderedDict
import json
import os


class Page(OrderedDict):
    """
    Represents a page belonging to a site in Hub. A Page is a layout of
    content that can be rendered within the context of a Site
    """

    def __init__(self, gis, page_item: Item):
        """
        Constructs an empty Page object
        """
        self.item = page_item
        self._gis = gis
        try:
            self._pagedict = self.item.get_data()
            pmap = PropertyMap(self._pagedict)
            self.definition = pmap
        except:
            self.definition = None

    def __repr__(self):
        return '<%s title:"%s" owner:%s>' % (
            type(self).__name__,
            self.item.title,
            self.item.owner,
        )

    @property
    def itemid(self) -> str:
        """
        Returns the item id of the page item
        """
        return self.item.id

    @property
    def title(self) -> str:
        """
        Returns the title of the page item
        """
        return self.item.title

    @property
    def description(self) -> str:
        """
        Getter/Setter for the page description
        """
        return self.item.description

    @description.setter
    def description(self, value):
        self.item.description = value

    @property
    def layout(self) -> InsensitiveDict:
        """
        Return layout of a page
        """
        return InsensitiveDict(self.definition["values"]["layout"])

    @property
    def owner(self) -> str:
        """
        Returns the owner of the page item
        """
        return self.item.owner

    @property
    def tags(self) -> str:
        """
        Returns the tags of the page item
        """
        return self.item.tags

    @tags.setter
    def tags(self, value):
        self.item.tags = value

    @property
    def slug(self) -> str:
        """
        Returns the page slug
        """
        return self.title.replace(" ", "-").lower()

    def update(self, page_properties=None, slug=None):
        """Updates the page.

        .. note::

            For page_properties, pass in arguments for only the properties you want to be updated.
            All other properties will be untouched.  For example, if you want to update only the
            page's description, then only provide the description argument in page_properties.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        page_properties           Required dictionary. See URL below for the keys and values.
        ---------------------     --------------------------------------------------------------------
        slug                      Optional string. The slug or subdomain for the page.
        =====================     ====================================================================

        To find the list of applicable options for argument page_properties -
        https://developers.arcgis.com/python/api-reference/arcgis.gis.toc.html#arcgis.gis.Item.update

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

            USAGE EXAMPLE: Update a page successfully

            page1 = mySite.pages.get('itemId12345')
            page1.update(page_properties={'description':'Description for page.'})

            >> True
        """
        _page_data = self.definition
        if page_properties:
            for key, value in page_properties.items():
                _page_data[key] = value
        if slug:
            # Fetch all the sites this page is connected to
            linked_sites = self.definition["values"]["sites"]
            for item in linked_sites:
                # Update the title and slug on the parent sites
                if self._gis._portal.is_arcgisonline:
                    site = self._gis.hub.sites.get(item["id"])
                else:
                    site = self._gis.sites.get(item["id"])
                site.definition["values"]["pages"] = [
                    p
                    for p in site.definition["values"]["pages"]
                    if p["id"] != self.itemid
                ]
                _renamed_page = {}
                _renamed_page["id"] = self.itemid
                _renamed_page["title"] = slug
                _renamed_page["slug"] = slug
                site.definition["values"]["pages"].append(_renamed_page)
                site.item.update(item_properties={"text": site.definition})
            # Update the slug on the page
            _page_data["title"] = slug
        return self.item.update(_page_data)

    def update_layout(self, layout: dict) -> bool:
        """Updates the layout of the page.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        layout                    Required dictionary. The new layout dictionary to update to the page.
        =====================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).
        .. code-block:: python
            USAGE EXAMPLE: Update a site successfully
            page1 = myHub.pages.get('itemId12345')
            page_layout = page1.layout
            page_layout.sections[0].rows[0].cards.pop(0)
            page1.update_layout(layout = page_layout)
            >> True
        """
        # Deleting the draft file for this page, if exists
        resources = self.item.resources.list()
        for resource in resources:
            if "draft-" in resource["resource"]:
                self.item.resources.remove(file=resource["resource"])
        # Update the data of the page
        self.definition["values"]["layout"] = layout._json()
        return self.item.update(item_properties={"text": self.definition})

    def delete(self) -> bool:
        """
        Deletes the page. If unable to delete, raises a RuntimeException.

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Delete a page successfully

            page1 = myHub.pages.get('itemId12345')
            page1.delete()

            >> True
        """
        # Unlink sites
        linked_sites = self.definition["values"]["sites"]
        for item in linked_sites:
            site_item = self._gis.content.get(item["id"])
            if self._gis._portal.is_arcgisonline:
                site = self._gis.hub.sites.get(item["id"])
            else:
                site = self._gis.sites.get(item["id"])
            definition = site.definition
            definition["values"]["pages"] = [
                p for p in site.definition["values"]["pages"] if p["id"] != self.itemid
            ]
            site_item.update(item_properties={"text": definition})
        # Remove delete protection on page
        self.item.protect(enable=False)
        # Delete page item
        self.item.delete()


class PageManager(object):
    """
    Helper class for managing pages within a Hub. This class is not created by users directly.
    An instance of this class, called 'pages', is available as a property of the Site object. Users
    call methods on this 'pages' object to manipulate (add, get, search, etc) pages for a site.
    """

    def __init__(self, gis, site=None):
        try:
            gis._portal
            self._gis = gis
            # For Hub (ArcGIS Online) sites
            if self._gis._portal.is_arcgisonline:
                self._hub = self._gis.hub
        # If accessed from the hub route for cloning from Enterprise to Hub
        except:
            self._hub = gis
            self._gis = self._hub.gis
        self._site = site

    def add(self, title: str, site=None) -> Page:
        """
        Returns the pages linked to the specific site.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        title                      Required string. The title of the new page.
        -----------------------    -------------------------------------------------------------
        site                       Optional Site. The site object to add the page to.
        =======================    =============================================================

        :return:
           The page if successfully added, None if unsuccessful.

        .. code-block:: python

            USAGE EXAMPLE: Add a page to a site successfully

            page1 = mySite.pages.add(title='My first page')
            page1.item

        .. code-block:: python

            USAGE EXAMPLE: Add a page successfully

            page2 = myHub.pages.add(title='My second page', site=mySite)
            page2.item
        """
        # If site object is not provided
        if site is None:
            if self._site is None:
                raise Exception("Site object needed for adding page")
        # If called from a specified site
        if self._site is not None:
            site = self._site
        # Fetch site collab group, if exists
        try:
            collab_group = self._gis.groups.get(site.collab_group_id)
        except:
            collab_group = None

        # For pages in ArcGIS Online
        if self._gis._portal.is_arcgisonline:
            # Set item details
            item_type = "Hub Page"
            typekeywords = "Hub, hubPage, JavaScript, Map, Mapping Site, Online Map, OpenData, selfConfigured, Web Map"
            description = "DO NOT DELETE OR MODIFY THIS ITEM. This item is managed by the ArcGIS Hub application. To make changes to this site, please visit https://hub.arcgis.com/overview/edit"
            image_card_url = "https://cloud.githubusercontent.com/assets/7389593/20107607/1d2c3844-a5a7-11e6-9ec0-9e389033ccd8.jpg"
        # For Enterprise Sites
        else:
            item_type = "Site Page"
            typekeywords = "Hub, hubPage, JavaScript, Map, Mapping Site, Online Map, OpenData, selfConfigured, Web Map"
            description = (
                "DO NOT DELETE OR MODIFY THIS ITEM. This item is managed by the ArcGIS Enterprise Sites application. To make changes to this site, please visit"
                + self._gis.url
                + "/apps/sites/#/home/overview/edit/"
            )
            image_card_url = (
                self._gis.url
                + "/apps/sites/images/placeholders/page-editor-card-image-placeholder.jpg"
            )
        # Create page item
        _item_dict = {
            "title": title,
            "type": item_type,
            "typeKeywords": typekeywords,
            "description": description,
            "culture": self._gis.properties.user.culture,
        }
        item = self._gis.content.add(_item_dict, owner=self._gis.users.me.username)

        # share page with content and core team groups
        if collab_group:
            item.share(groups=[collab_group])

        # protect page from accidental deletion
        item.protect(enable=True)

        # Fetching page data
        data_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "_store/pages-data.json")
        )
        with open(data_path) as f:
            _page_data = json.load(f)

        # Updating page data
        _page_data["values"]["layout"]["sections"][1]["rows"][0]["cards"][0][
            "component"
        ]["settings"]["src"] = image_card_url
        _page_data["values"]["updatedBy"] = self._gis.users.me.username
        _data = json.dumps(_page_data)
        item.update(item_properties={"text": _data})
        page = Page(self._gis, item)
        # Link page to site
        status = site.pages.link(page)
        if status:
            return page

    def clone(self, page: Page, site=None) -> Page:
        """
        Clone allows for the creation of a page that is derived from the current page.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        page                Required Page object of page to be cloned.
        ---------------     --------------------------------------------------------------------
        site                Optional Site object.
        ===============     ====================================================================

        :return:
           Page.

        .. code-block:: python

            USAGE EXAMPLE: Clone a page to a site successfully

            #Fetch page
            page1 = myhub.pages.get('itemId12345')
            page1.item

            #Clone it
            page2 = mySite.pages.clone(page1)
            page2.item
        """
        from datetime import timezone

        now = datetime.now(timezone.utc)
        # New title
        title = page.title + "-copy-%s" % int(now.timestamp() * 1000)
        # Checking if item of correct type has been passed
        if "hubPage" not in page.item.typeKeywords:
            raise Exception("Incorrect item type. Page item needed for cloning.")
        # If site object is not provided
        if site is None:
            if self._site is None:
                raise Exception("Site object needed for cloning page")
        # If called from a specified site
        if self._site is not None:
            site = self._site
        # Create new page within site
        _cloned_page = site.pages.add(title)
        # Copy the page layout
        _cloned_page.definition["values"]["layout"] = page.definition["values"][
            "layout"
        ]
        _cloned_page.item.update(item_properties={"text": _cloned_page.definition})
        return Page(self._gis, _cloned_page.item)

    def get(self, page_id: str) -> Page:
        """
        Returns the page object for the specified page_id.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        page_id                    Required string. The page itemid.
        =======================    =============================================================

        :return:
            The page object if the item is found, None if the item is not found.

        .. code-block:: python

            USAGE EXAMPLE: Fetch a page successfully
            page1 = myHub.pages.get('itemId12345')
            page1.item
        """
        pageItem = self._gis.content.get(page_id)
        if "hubPage" in pageItem.typeKeywords:
            return Page(self._gis, pageItem)
        else:
            raise TypeError("Item is not a valid page or is inaccessible.")

    def link(self, page, site=None, slug=None):
        """
        Links the page to the specific site.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        page                       Required string. The page object to link.
        -----------------------    -------------------------------------------------------------
        site                       Optional Site. The site object to link page to.
        -----------------------    -------------------------------------------------------------
        slug                       Optional string. The slug reference of the page in this site.
        =======================    =============================================================

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Link a page successfully for specific site
            mySite.pages.link(page_id='itemId12345')

            >> True

        .. code-block:: python

            USAGE EXAMPLE: Link a page successfully for site object passed as param
            myHub.pages.link(page_id='itemId12345', site=mySite)

            >> True
        """
        _new_page = {}
        _new_site = {}
        # Checking if item of correct type has been passed
        if "hubPage" not in page.item.typeKeywords:
            raise Exception("Incorrect item type. Page item needed for cloning.")
        # If site object is not provided
        if site is None:
            if self._site is None:
                raise Exception("Site object needed for linking page")
        # If called from a specified site
        if self._site is not None:
            site = self._site
        # Create new page dictionary
        _site_data = site.definition
        _new_page["id"] = page.itemid
        _new_page["title"] = page.title
        if slug is not None:
            _new_page["slug"] = slug.replace(" ", "-").lower()
        else:
            _new_page["slug"] = page.slug
        _site_data["values"]["pages"].append(_new_page)
        # Create new site dictionary
        _page_data = page.definition
        _new_site["id"] = site.itemid
        _new_site["title"] = site.title
        _page_data["values"]["sites"].append(_new_site)
        # Update page and site data with new linking
        page.item.update(item_properties={"text": _page_data})
        return site.item.update(item_properties={"text": _site_data})

    def unlink(self, page: Page, site=None) -> bool:
        """
        Unlinks the page from the specific site.

        .. note::

            This method will only detach the page from the site. In order to remove links to this
            page from the site's UI, you will have to manually remove it from the site through
            the site's Layout Editior.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        page                       Required Page. The page object to unlink.
        -----------------------    -------------------------------------------------------------
        site                       Optional Site. The site object to unlink page from.
        =======================    =============================================================

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Unlink a page successfully from specific site

            mySite.pages.unlink(page_id='itemId12345')

            >> True

        .. code-block:: python

            USAGE EXAMPLE: Unlink a page successfully from site object passed as param

            myHub.pages.unlink(page_id='itemId12345', site=mySite)

            >> True
        """
        # If site object is not provided
        if site is None:
            if self._site is None:
                raise Exception("Site object needed for unlinking page")
        # Checking if item of correct type has been passed
        if "hubPage" not in page.item.typeKeywords:
            raise Exception("Incorrect item type. Page item needed for cloning.")
        # If called from a specified site
        if self._site is not None:
            site = self._site
        # Update site and pages with the unlinking
        _site_data = site.definition
        _page_data = page.definition
        _site_data["values"]["pages"] = [
            p for p in _site_data["values"]["pages"] if p["id"] != page.itemid
        ]
        _page_data["values"]["sites"] = [
            s for s in _page_data["values"]["sites"] if s["id"] != site.itemid
        ]
        # Delete page if it has no other site linkage
        if len(_page_data["values"]["sites"]) == 0:
            page.delete()
        # Update page data to reflect unlinking
        else:
            page.item.update(item_properties={"text": _page_data})
        # Update site data to reflect unlinking
        return site.item.update(item_properties={"text": _site_data})

    def search(self, title=None, owner=None, created=None, modified=None, tags=None):
        """
        Searches for pages.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. Return pages with provided string in title.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. Return pages owned by a username.
        ---------------     --------------------------------------------------------------------
        created             Optional string. Date the page was created.
                            Shown in milliseconds since UNIX epoch.
        ---------------     --------------------------------------------------------------------
        modified            Optional string. Date the page was last modified.
                            Shown in milliseconds since UNIX epoch
        ---------------     --------------------------------------------------------------------
        tags                Optional string. User-defined tags that describe the page.
        ===============     ====================================================================

        :return:
           A list of matching pages.
        """

        pagelist = []

        if self._site is not None:
            pages = self._site.definition["values"]["pages"]
            items = [self._gis.content.get(page["id"]) for page in pages]
        # Build search query
        else:
            query = "typekeywords:hubPage"
            if title != None:
                query += " AND title:" + title
            if owner != None:
                query += " AND owner:" + owner
            if created != None:
                query += " AND created:" + created
            if modified != None:
                query += " AND modified:" + modified
            if tags != None:
                query += " AND tags:" + tags

            # Search
            items = self._gis.content.search(query=query, max_items=5000)

        # Return searched pages
        for item in items:
            pagelist.append(Page(self._gis, item))
        return pagelist



import glob
import math
import os
import random
import traceback
import types
from itertools import islice
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from .._utils.env import raise_fastai_import_error

HAS_FASTAI = True

try:
    import torch
    from .._utils.common import ArcGISMSImage
    from ._base_data import ArcgisData
    from .._utils.pointcloud_data import get_device
    from fastai.basic_data import DatasetType
    from fastai.vision.data import imagenet_stats
    from fastai.core import subplots
    from fastai.data_block import DataBunch
    from fastai.vision.data import ImageDataBunch
    from fastai.torch_core import grab_idx, to_detach
    import torchvision.transforms.functional as TF
    from typing import Sequence
    from fastai.vision.image import Image, pil2tensor
    from numpy import ma
    from PIL import Image as PILImage
    from scipy.ndimage.morphology import distance_transform_edt
    from torch import Tensor
    from torch.utils.data import DataLoader, Dataset
    from torchvision import transforms as pytorch_tfms
except ImportError as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


def _to_np(x: Tensor) -> np.ndarray:
    x = x.numpy()
    if x.ndim > 2:
        x = x * 255.0
        return x.transpose(1, 2, 0).astype(np.uint8)
    return x


def _show_batch(
    self,
    rows: int = 5,
    ds_type: DatasetType = DatasetType.Train,
    cpu: bool = True,
    **kwargs,
):
    "Show a batch of data in `ds_type` on a few `rows`."
    if self is None:
        return

    dl = self.dl(ds_type)
    w = dl.num_workers
    dl.num_workers = 0
    try:
        x, labels = next(iter(dl))
    finally:
        dl.num_workers = w

    y, o = labels
    x, y, o = to_detach(x, cpu=cpu), to_detach(y, cpu=cpu), to_detach(o, cpu=cpu)
    y, o = y.squeeze(1), o.squeeze(1)

    n_items = min(self.dl(ds_type).batch_size, rows, x.size(0))
    xs = [_to_np(grab_idx(x, i)) for i in range(n_items)]
    ys = [_to_np(grab_idx(y, i)) for i in range(n_items)]
    orients = [_to_np(grab_idx(o, i)) for i in range(n_items)]
    _show_pairs(xs, ys, orients, bin_size=self.orient_bin_size)


def _show_pairs(
    xs,
    ys,
    orients,
    imgsize: int = 4,
    figsize: Optional[Tuple[int, int]] = None,
    bin_size: int = 20,
):
    """
    Show Image - Road Label - Orientation Vector label pairs
    """
    rows = len(xs)
    main_title = "Image / Binary Road Maks"
    axs = subplots(rows, 2, imgsize=imgsize, figsize=figsize, title=main_title)
    for x, y, o, ax in zip(xs, ys, orients, axs):
        ax[0].imshow(x)
        ax[1].imshow(y)
    for ax in axs.flatten():
        ax.axis("off")


def _plotOrientationOnImage(ax, orientMap, image, bin_size=20):
    """
    Plot Orientation Vectors overlay on Image
    """
    from .._utils.road_orient_utils.affinity_utils import convertAngles2VecMap

    ax.imshow(image)
    orientmap_xy = convertAngles2VecMap(orientMap.shape, orientMap, bin_size)
    U = orientmap_xy[:, :, 0] * -1
    V = orientmap_xy[:, :, 1]
    X, Y = np.meshgrid(np.arange(U.shape[0]), np.arange(U.shape[1]))
    M = np.zeros(U.shape, dtype="bool")
    M[U**2 + V**2 < 0.5 * 0.5] = True
    U = ma.masked_array(U, mask=M)
    V = ma.masked_array(V, mask=M)

    s = 15
    ax.quiver(
        X[::s, ::s],
        Y[::s, ::s],
        U[::s, ::s],
        V[::s, ::s],
        scale=50,
        headaxislength=3,
        headwidth=4,
        width=0.01,
        alpha=0.8,
        color="r",
    )


class DiscreteAffine:
    """
    Custom PyTorch Rotation transform.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    angles                   Required List of int ranging from 0..360
    ---------------------   -------------------------------------------
    """

    def __init__(self, angles: Sequence[int]):
        self.angles = angles

    def __call__(self, x):
        angle = random.choice(self.angles)
        return TF.affine(x, angle, (0, 0), 1.0, 0.0, resample=False, fillcolor=0)


class RoadOrientation:
    def __init__(self, orig_data, classified_tiles_data: ArcgisData, **kwargs):
        """
        The class is used to create Fast AI Databunch for Multi-Task
        Road - Orientation learning. It internally utilizes RoadOrientation
        Dataset.

        Reference Paper: CVPR 2019 - Improved Road Connectivity by Joint
                        Learning of Orientation and Segmentation

        Parameters
        --------------------------------------------------------------------
        @param classified_tiles_data: instance of
            arcgis.learn._data_utils._pixel_classifier_data.ClassifiedTilesData
        @param kwargs:
        |--> orient_bin_size: float, Bin size used to quantize Orientation angles
        |--> orient_theta: float, number used as width of orientation angles
        |--> multi_scale: list of floats [0..1], used to create multi-scale label masks
        |                    e.g. [0.25,0.5, 1] will return 3 masks with dimensions
        |                    256x256, 512x512, 1024x1024 if the image size is 1024x1024
        """
        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception)

        self.base = classified_tiles_data
        assert (
            self.base
        ), "Road Orientation requires instance of base Classified Tiles Dataset!"
        # TODO: Need to fix this Hack
        if "0" not in self.base.class_mapping:
            self.base.class_mapping[0] = "0"
        self.base.class_mapping = dict(sorted(self.base.class_mapping.items()))

        self.class_mapping = (
            {"0": 0, "1": 1}
            if not bool(self.base.class_mapping) is None
            else self.base.class_mapping
        )
        # Dataset parameters
        default_road_params = {
            "orient_bin_size": 20,
            "orient_theta": 8.0,
            "multi_scale": None,
        }
        # self.road_extractor_params = kwargs.get("road_extractor_params", default_road_params)
        self.orient_bin_size = kwargs.get(
            "orient_bin_size", default_road_params["orient_bin_size"]
        )
        self.orient_theta = kwargs.get(
            "orient_theta", default_road_params["orient_theta"]
        )
        self.multi_scale = kwargs.get("multi_scale", default_road_params["multi_scale"])

        self.files = []
        for cnt, image in enumerate(orig_data.train_ds.x.items):
            file_pair = {
                "image": image,
                "label": orig_data.train_ds.y.items[cnt],
            }
            self.files.append(file_pair)

        # self.valid_files = []
        for cnt, image in enumerate(orig_data.valid_ds.x.items):
            file_pair = {
                "image": image,
                "label": orig_data.valid_ds.y.items[cnt],
            }
            self.files.append(file_pair)

        random.shuffle(self.files)
        if self.base.data_count:
            print(
                f"Taking only {self.base.data_count} samples from total files {len(self.files)}"
            )
            self.files = self.files[: self.base.data_count]
            print(f"New Data Pairs => {len(self.files)}")
        files_iterator = iter(self.files)
        val_split_count = int(len(self.files) * self.base.val_split_pct)
        self.valid_files = list(islice(files_iterator, val_split_count))
        self.train_files = list(files_iterator)
        self.transforms = (
            [
                # Training Transforms
                [
                    # Pairwise Transforms
                    [
                        pytorch_tfms.RandomCrop(
                            size=self.base.chip_size,
                            padding_mode="constant",
                            pad_if_needed=True,
                        )
                        if self.base.chip_size
                        else None,
                        pytorch_tfms.Resize(size=self.base.resize_to)
                        if self.base.resize_to
                        else None,
                        pytorch_tfms.RandomHorizontalFlip(),
                        pytorch_tfms.RandomVerticalFlip(),
                        pytorch_tfms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        ),
                    ],
                    # Image Transforms
                    [
                        pytorch_tfms.ColorJitter(brightness=0.1, contrast=0.1),
                        pytorch_tfms.ToTensor(),
                    ],
                ],
                # Validation Transforms
                [
                    # Pairwise Transforms
                    [
                        pytorch_tfms.Resize(size=self.base.resize_to)
                        if self.base.resize_to
                        else None
                    ],
                    # Image Transforms
                    [pytorch_tfms.ToTensor()],
                    [
                        pytorch_tfms.Normalize(
                            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                        )
                    ],
                ],
            ]
            if self.base.transforms is None
            else self.base.transforms
        )

    def get_databunch(self, orig_data, **kwargs) -> ImageDataBunch:
        """
        Method to create Databunch
        """
        train_dataset = RoadOrientDataset(
            orig_data,
            self.train_files,
            self.transforms[0],
            **kwargs,
        )
        valid_dataset = RoadOrientDataset(
            orig_data,
            self.valid_files,
            self.transforms[1],
            **kwargs,
        )
        train_dl = DataLoader(
            train_dataset,
            batch_size=self.base.train_batch_size,
            **self.base.databunch_kwargs,
        )
        valid_dl = DataLoader(
            valid_dataset,
            batch_size=self.base.valid_batch_size,
            **self.base.databunch_kwargs,
        )

        device = get_device()
        data = ImageDataBunch(train_dl, valid_dl, device=device)

        data.chip_size = data.train_ds[0][0].shape[-1]
        data.c = len(self.class_mapping) if self.class_mapping else 2

        data.orient_c = int(360.0 / self.orient_bin_size) + 1
        data.orient_bin_size = self.orient_bin_size
        data.orient_theta = self.orient_theta
        data.multi_scale = self.multi_scale
        data.classes = []
        for _, value in self.class_mapping.items():
            if value == 0:
                data.classes.append("NoData")
                continue
            data.classes.append(value)
        data.extension = self.base.extension
        data.val_files = self.valid_files
        data.show_batch = types.MethodType(_show_batch, data)
        data.parent_obj = self
        data.sub_dataset_type = "RoadOrientation"
        data.transform = self.transforms
        x_shape = data.train_ds[0][0].shape

        return data


class RoadOrientDataset(Dataset):
    def __init__(
        self,
        orig_data,
        data_files: Dict,
        transforms: List = None,
        multi_scale: List = None,
        orient_theta: int = 5,
        orient_bin_size: int = 20,
        **kwargs,
    ):
        """
        PyTorch Dataset class to create Road-Orientaion pair for training.

        @param data_files: Dict, contain image-label pair file paths
        @param transforms: list of Fastai transforms
        @param multi_scale: list of floats [0..1], used to create multi-scale label masks
                            e.g. [0.25,0.5, 1] will return 3 masks with dimensions
                            256x256, 512x512, 1024x1024 if the image size is 1024x1024
        @param orient_theta: float, number used as width of orientation angles
        @param orient_bin_size: float, Bin size used to quantize Orientation angles
        @param kwargs:
            |--> is_gaussian_mask: bool, if True code assumes that given road labels are
            |                      gaussian and do not convert binary road mask to gaussian roads
            |--> gaussian_thresh: float, used to change the road width using gaussian mask
            |--> generate_orient: bool, flag will allow to create Orientation Label. It is
            |                     helpful at inference time for faster data prepration.
        """
        self.data = kwargs.get("data", None)
        self.orig_data = orig_data
        self.files = data_files
        self.pair_tfms = [tfm for tfm in transforms[0] if tfm]
        self.image_tfms = [tfm for tfm in transforms[1] if tfm]
        self.pair_tfms = (
            pytorch_tfms.Compose(self.pair_tfms) if self.pair_tfms else None
        )
        self.image_tfms = (
            pytorch_tfms.Compose(self.image_tfms) if self.image_tfms else None
        )

        self.gaussian_thresh = kwargs.get("gaussian_thresh", 0.76)
        self.is_gaussian_mask = kwargs.get("is_gaussian_mask", False)
        self.generate_orient = kwargs.get("generate_orient", True)

        # Angle Mask Buffers
        self.angle_theta = orient_theta
        self.bin_size = orient_bin_size

    def __len__(self):
        return len(self.files)

    def __getitem__(self, index):
        image, label = self._getRoadData(index)

        seed = np.random.randint(2147483647)  # make a seed with numpy generator
        random.seed(seed)  # apply this seed to img tranfsorms
        label = np.asarray(label)
        if isinstance(image, PILImage.Image):
            image = torch.from_numpy(
                np.asarray(image).astype(np.float32).transpose(2, 0, 1)
            )
        if self.generate_orient:
            orient_label = self._getOrientationGT(np.copy(label.astype(np.uint8)))
            # orient_label = self._get_fastai_image(orient_label, dtype=np.float32)
            return (
                image,
                [
                    torch.from_numpy(label.copy()),
                    torch.from_numpy(orient_label.copy()),
                ],
            )
        else:
            return (
                image,
                [
                    torch.from_numpy(label.copy()),
                    torch.from_numpy(label.copy()),
                ],
            )

    def _get_fastai_image(self, x, dtype) -> Image:
        """
        Convert PILImage or np array to Fastai.Vision.Image
        """
        if isinstance(x, PILImage.Image):
            tensor = pil2tensor(x, dtype=dtype)
        elif isinstance(x, np.ndarray):
            tensor = torch.from_numpy(x.astype(dtype))
        return Image(tensor)

    def _getRoadData(self, index: int) -> Tuple:
        """
        Get Sattellite Image and Road Label Mask
        """
        image_dict = self.files[index]
        # read each image in list
        image = ArcGISMSImage.open(image_dict["image"])
        if (image.shape[0] > 3) and (self.orig_data._imagery_type == "RGB"):
            image = image.data[[0, 1, 2]]
        label = self._get_mask(ArcGISMSImage.read_image(image_dict["label"]))

        return image, PILImage.fromarray(label).convert("L")

    def _getOrientationGT(self, road_mask: np.ndarray) -> np.ndarray:
        """
        Create Orientation Label for given road label
        """
        from .._utils.road_orient_utils.affinity_utils import (
            getKeypoints,
            getVectorMapsAngles,
        )

        height, width = (
            road_mask.shape if isinstance(road_mask, np.ndarray) else road_mask.size
        )
        smooth_dist_dict = {6: 1, 7: 1, 8: 1, 9: 2, 10: 4, 11: 4}
        keypoints = getKeypoints(
            road_mask.astype(np.float32),
            is_gaussian=False,
            thresh=0.98,
            smooth_dist=smooth_dist_dict[round(math.log(height) / math.log(2))],
        )
        _, vecmap_angles = getVectorMapsAngles(
            (height, width), keypoints, theta=self.angle_theta, bin_size=self.bin_size
        )

        return vecmap_angles  # np.expand_dims(vecmap_angles, 0)

    def _createGaussianMask(self, gt_array):
        """
        Create Gaussian road mask for given road label.
            1 - Perform Skeletonization
            2 - Compute Distance Transform
            3 - Convert Gaussian Road Mask using standard Deviation of 15
        """
        from skimage.morphology import skeletonize

        gt_array = skeletonize(gt_array)
        distance_array = distance_transform_edt(1 - (gt_array))
        std = 15
        distance_array = np.exp(-0.5 * (distance_array * distance_array) / (std * std))

        return distance_array

    def _get_mask(self, label: np.ndarray) -> np.ndarray:
        """
        Threshold the gaussian mask.
        """
        from skimage import filters

        if self.is_gaussian_mask:
            new_label = (np.array(label).astype(float)) / 255.0
        else:
            try:
                new_label = self._createGaussianMask(label)
            except:
                new_label = self._createGaussianMask(
                    label > filters.threshold_otsu(label)
                )
        new_label[new_label >= self.gaussian_thresh] = 1
        new_label[new_label < self.gaussian_thresh] = 0

        return new_label.astype(np.uint8)



import copy
from functools import partial
from pathlib import Path
import sys
import json
import warnings
import traceback
from ..models._arcgis_model import ArcGISModel, model_characteristics_folder
from .._utils._shap_masker import custom_tokenizer

HAS_NUMPY = True
HAS_FASTAI = True

try:
    import torch
    import torch.nn as nn
    import pandas as pd
    from fastai.text.transform import Tokenizer
    from fastprogress.fastprogress import progress_bar
    from fastai.basic_train import Learner, DatasetType
    from fastai.train import to_fp16
    from fastai.metrics import accuracy, error_rate, accuracy_thresh

    # from transformers import AdamW
    from transformers import AutoTokenizer, AutoConfig
    from .._utils.env import _LAMBDA_TEXT_CLASSIFICATION

    if not _LAMBDA_TEXT_CLASSIFICATION:
        from sklearn.metrics import classification_report
    from ._arcgis_transformer import ModelBackbone, infer_model_type
    from .._utils.common import _get_emd_path
    from .._utils.text_data import (
        TextDataObject,
        save_data_in_model_metrics_html,
        copy_metrics,
    )
    from .._utils.text_transforms import TransformersBaseTokenizer, TransformersVocab
    from ._transformer_text_classifier import (
        TransformerForTextClassification,
        backbone_models_reverse_map,
        transformer_architectures,
        transformer_seq_length,
    )
    from transformers import logging

    logging.get_logger("filelock").setLevel(logging.ERROR)
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False
    from ._transformer_text_classifier import (
        transformer_architectures,
        transformer_seq_length,
    )

    class TransformerForTextClassification:
        _supported_backbones = transformer_architectures

else:
    warnings.filterwarnings("ignore", category=UserWarning, module="fastai")

try:
    import numpy as np

    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
except:
    HAS_NUMPY = False


warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")


class TextClassifier(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.text.TextClassifier` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Optional data object returned from :class:`~arcgis.learn.prepare_textdata` function.
                            data object can be `None`, in case where someone wants to use a
                            Hugging Face Transformer model fine-tuned on classification task.
                            In this case the model should be used directly for inference.
    ---------------------   -------------------------------------------
    backbone                Optional string. Specifying the HuggingFace
                            transformer model name to be used to train the
                            classifier. Default set to `bert-base-cased`.

                            To learn more about the available models or
                            choose models that are suitable for your dataset,
                            kindly visit:- https://huggingface.co/transformers/pretrained_models.html

                            To learn more about the available transformer models fine-tuned
                            on Text Classification Task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=text-classification
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    verbose                 Optional string. Default set to `error`. The
                            log level you want to set. It means the amount
                            of information you want to display while training
                            or calling the various methods of this class.
                            Allowed values are - `debug`, `info`, `warning`,
                            `error` and `critical`.
    ---------------------   -------------------------------------------
    seq_len                 Optional Integer. Default set to 512. Maximum
                            sequence length (at sub-word level after tokenization)
                            of the training data to be considered for training
                            the model.
    ---------------------   -------------------------------------------
    thresh                  Optional Float. This parameter is used to set
                            the threshold value to pick labels in case of
                            multi-label text classification problem. Default
                            value is set to 0.25
    ---------------------   -------------------------------------------
    mixed_precision         Optional Bool. Default set to False. If set
                            True, then mixed precision training is used
                            to train the model
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.TextClassifier` Object
    """

    # supported transformer backbones
    supported_backbones = transformer_architectures

    def __init__(self, data, backbone="bert-base-cased", **kwargs):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)

        self.logger = logging.get_logger()
        if kwargs.get("verbose", None):
            self.logger.setLevel(kwargs.get("verbose").upper())
        else:
            self.logger.setLevel(logging.ERROR)

        model_backbone = ModelBackbone(backbone)
        super().__init__(data, model_backbone)
        self._emodel = None
        self._emask = None
        self.shap_values = None
        self.is_multilabel_problem = False
        self.thresh = kwargs.get("thresh", 0.25)
        self._mixed_precision = kwargs.get("mixed_precision", False)
        self._seq_len = kwargs.get("seq_len", transformer_seq_length)
        model_config = kwargs.get("model_config", None)
        if data is None:
            model = TextClassifier.from_pretrained(backbone, **kwargs)
            self.learn = model.learn
            self._data = model._data
        else:
            self._create_text_learner_object(
                data,
                backbone,
                kwargs.get("pretrained_path", None),
                config=model_config,
                mixed_precision=self._mixed_precision,
                seq_len=self._seq_len,
            )

            self.learn.model = self.learn.model.to(self._device)
            layer_groups = self.learn.model.get_layer_groups()
            self.learn.split(layer_groups)
            # self._freeze()

    def _create_text_learner_object(
        self,
        data,
        backbone,
        pretrained_path=None,
        mixed_precision=False,
        seq_len=transformer_seq_length,
        config=None,
    ):
        model_type = infer_model_type(backbone, transformer_architectures)
        self.logger.info(f"Inferred Backbone: {model_type}")
        pretrained_model_name = backbone

        if not config:
            config = AutoConfig.from_pretrained(pretrained_model_name)
        transformer_tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name, config=config
        )

        # pad_first = bool(model_type in ['xlnet'])
        pad_first = True if transformer_tokenizer.padding_side == "left" else False
        pad_idx = transformer_tokenizer.pad_token_id

        base_tokenizer = TransformersBaseTokenizer(
            pretrained_tokenizer=transformer_tokenizer, seq_len=seq_len
        )
        tokenizer = Tokenizer(tok_func=base_tokenizer, pre_rules=[], post_rules=[])
        if sys.platform == "win32":
            tokenizer.n_cpus = 1
        vocab = TransformersVocab(tokenizer=transformer_tokenizer)

        if data._is_empty or data._backbone != backbone:
            self.logger.info("Creating DataBunch")
            classes = None
            data._prepare_databunch(
                tokenizer=tokenizer,
                vocab=vocab,
                pad_first=pad_first,
                pad_idx=pad_idx,
                backbone=backbone,
                classes=classes,
                logger=self.logger,
            )

        databunch = data.get_databunch()
        if config.label2id != databunch.train_ds.c2i:
            config.label2id = databunch.train_ds.c2i
            config.id2label = {y: x for x, y in config.label2id.items()}

        if pretrained_path is not None:
            pretrained_path = str(_get_emd_path(pretrained_path))

        model = TransformerForTextClassification(
            architecture=model_type,
            pretrained_model_name=pretrained_model_name,
            config=config,
            pretrained_model_path=pretrained_path,
            seq_len=seq_len,
        )
        model.init_model()
        # opt_func = partial(AdamW, correct_bias=False)

        self.is_multilabel_problem = True if len(data._label_cols) > 1 else False
        if self.is_multilabel_problem:
            accuracy_multi = partial(accuracy_thresh, thresh=self.thresh)
            accuracy_multi.__name__ = "accuracy"
            metrics = [accuracy_multi]
            loss_func = nn.BCEWithLogitsLoss()
            # self.learn = Learner(databunch, model, opt_func=opt_func, loss_func=loss_func, metrics=metrics)
            self.learn = Learner(
                databunch, model, loss_func=loss_func, metrics=metrics, path=data.path
            )
        else:
            metrics = [accuracy, error_rate]

            # self.learn = Learner(databunch, model, opt_func=opt_func, metrics=metrics)
            self.learn = Learner(databunch, model, metrics=metrics, path=data.path)

        if pretrained_path is not None:
            self.load(pretrained_path)

        if mixed_precision:
            if model_type in ["xlnet", "mobilebert"]:
                error_message = (
                    f"Mixed precision training is not supported for transformer model - {model_type.upper()}."
                    "\nKindly turn off the `mixed_precision` flag to use this model in its default mode,"
                    f" or choose a different transformer architectures from - {transformer_architectures}"
                )
                raise Exception(error_message)
            self.logger.info("Converting model to 16 Bit Floating Point precision")
            self.learn = to_fp16(self.learn)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy", "error_rate"]

    @classmethod
    def available_backbone_models(cls, architecture):
        """
        Get available models for the given transformer backbone

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        architecture            Required string. name of the transformer
                                backbone one wish to use. To learn more about
                                the available models or choose models that are
                                suitable for your dataset, kindly visit:-
                                https://huggingface.co/transformers/pretrained_models.html
        =====================   ===========================================

        :return: a tuple containing the available models for the given transformer backbone
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)
        return TransformerForTextClassification._available_backbone_models(architecture)

    def freeze(self):
        """
        Freeze up to last layer group to train only the last layer group of the model.
        """
        self.learn.freeze()

    @classmethod
    def from_pretrained(cls, backbone, **kwargs):
        """
        Creates an TextClassifier model object from an already fine-tuned
        Hugging Face Transformer backbone.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        backbone                Required string. Specify the Hugging Face Transformer
                                backbone name fine-tuned on Text Classification task.

                                To get more details on available transformer models
                                fine-tuned on Text Classification Task, kindly visit:-
                                https://huggingface.co/models?pipeline_tag=text-classification

        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.TextClassifier` Object
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)

        model_config = AutoConfig.from_pretrained(backbone)
        class_labels = list(model_config.id2label.values())
        data = TextDataObject(task="classification")
        data._backbone = backbone
        data.create_empty_object_for_classification("", [], class_labels)

        cls_object = cls(data, backbone, model_config=model_config, **kwargs)
        data.emd, data.emd_path = cls_object._get_emd_params(), None
        cls_object._data._is_empty = True
        return cls_object

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an TextClassifier model object from a Deep Learning
        Package(DLPK) or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :class:`~arcgis.learn.prepare_textdata` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.TextClassifier` model Object
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        pretrained_model = emd["PretrainedModel"]
        mixed_precision = emd["MixedPrecisionTraining"]
        text_cols = emd["TextColumns"]
        label_cols = emd["LabelColumns"]
        class_labels = list(emd["Label2Id"].keys())
        is_multilabel_problem = emd["IsMultilabelClassificationProblem"]
        thresh = emd.get("Threshold")
        seq_len = emd.get("SequenceLength", transformer_seq_length)

        data_is_none = False
        if data is None:
            data_is_none = True
            data = TextDataObject(task="classification")
            data._backbone = pretrained_model
            data.create_empty_object_for_classification(
                text_cols, label_cols, class_labels, is_multilabel_problem
            )
            data.emd, data.emd_path = emd, emd_path.parent
        cls_object = cls(
            data,
            pretrained_model,
            pretrained_path=str(emd_path),
            mixed_precision=mixed_precision,
            thresh=thresh,
            seq_len=seq_len,
        )
        if data_is_none:
            cls_object._data._is_empty = True
        return cls_object

    def load(self, name_or_path):
        """
        Loads a saved TextClassifier model from disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        if "\\" in str(name_or_path) or "/" in str(name_or_path):
            name_or_path = str(_get_emd_path(name_or_path))
        else:
            name_or_path = Path("models") / name_or_path
            name_or_path = str(_get_emd_path(name_or_path))
        return super().load(name_or_path)

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        compute_metrics=True,
        save_optimizer=False,
        **kwargs,
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by :class:`~arcgis.learn.SingleShotDetector`, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
                                Boolean `zip_files` if True, it will create the Deep
                                Learning Package (DLPK) file while saving the model.
        =====================   ===========================================

        :return: the qualified path at which the model is saved
        """
        from ..models._arcgis_model import _create_zip

        zip_files = kwargs.pop("zip_files", True)
        overwrite = kwargs.pop("overwrite", False)
        path = super().save(
            name_or_path,
            framework,
            publish=False,
            gis=None,
            compute_metrics=compute_metrics,
            save_optimizer=save_optimizer,
            zip_files=False,
            **kwargs,
        )

        self._save_df_to_html(path)

        if zip_files:
            _create_zip(path.name, str(path))

        if publish:
            self._publish_dlpk(
                (path / path.stem).with_suffix(".dlpk"), gis=gis, overwrite=overwrite
            )

        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io

        with io.capture_output() as captured:
            metrics = {"Accuracy": self.accuracy()}
            per_class_metric_df = self.metrics_per_label()
            metrics["MetricsPerLabel"] = json.dumps(
                per_class_metric_df.transpose().to_dict()
            )
        return metrics

    def _get_emd_params(self, save_inference_file=True):
        _emd_template = {}
        is_multilabel_problem = True if len(self._data._label_cols) > 1 else False
        _emd_template["Architecture"] = self.learn.model._transformer_architecture
        _emd_template[
            "PretrainedModel"
        ] = self.learn.model._transformer_pretrained_model_name
        _emd_template["ModelType"] = "Transformer"
        _emd_template["MixedPrecisionTraining"] = self._mixed_precision
        _emd_template["TextColumns"] = self._data._text_cols
        _emd_template["LabelColumns"] = self._data._label_cols
        _emd_template["Label2Id"] = self.learn.model._config.label2id
        _emd_template["SequenceLength"] = self._seq_len
        if is_multilabel_problem:
            _emd_template["Threshold"] = self.thresh
        _emd_template["IsMultilabelClassificationProblem"] = is_multilabel_problem
        return _emd_template

    def show_results(self, rows=5, **kwargs):
        """
        Prints the rows of the dataframe with target and prediction columns.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================

        :return: dataframe
        """
        self._check_requisites()
        if kwargs.get("thresh") is None and self.is_multilabel_problem:
            kwargs.update({"thresh": self.thresh})
        return self.learn.show_results(rows=rows, **kwargs)

    def accuracy(self):
        """
        Calculates the following  metric:

        * accuracy:   the number of correctly predicted labels in the validation set divided by the total number of items in the validation set

        :return: a floating point number depicting the accuracy of the classification model.
        """
        try:
            self._check_requisites()
        except Exception as e:
            acc = self._data.emd.get("Accuracy")
            if acc:
                return acc
            else:
                self.logger.error("Metric not found in the loaded model")
        else:
            if not HAS_NUMPY:
                self.logger.error("This function requires numpy.")
                return
            if hasattr(self.learn, "recorder"):
                metrics_names = self.learn.recorder.metrics_names
                metrics_values = self.learn.recorder.metrics
                if len(metrics_names) > 0 and len(metrics_values) > 0:
                    metrics = {
                        x: round(metrics_values[-1][i].item(), 4)
                        for i, x in enumerate(metrics_names)
                    }
                    metric = metrics["accuracy"]
                else:
                    metric = self._calculate_model_metric()
            else:
                metric = self._calculate_model_metric()
            return metric

    def _calculate_model_metric(self):
        self.logger.info("Calculating Model Metrics")
        validation_dataframe = self._data._valid_df

        if self.is_multilabel_problem:
            predictions = [
                x[2]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist()
                )
            ]
            labels = [
                [int(getattr(item, column)) for column in self._data._label_cols]
                for idx, item in validation_dataframe.iterrows()
            ]

            metric = accuracy_thresh(
                torch.tensor(predictions),
                torch.tensor(labels),
                thresh=self.thresh,
                sigmoid=False,
            ).item()
        else:
            predictions = [
                x[1]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist()
                )
            ]
            labels = [x[0] for x in validation_dataframe[self._data._label_cols].values]
            metric = round((np.sum(np.array(predictions) == labels) / len(labels)), 4)

        return metric

    def _predict(self, text, thresh=None):
        if thresh is None:
            thresh = self.thresh
        result = self.learn.model.predict_class(
            text, self._device, self.is_multilabel_problem, thresh
        )
        return result

    def _predict_batch(self, text, thresh=None):
        if thresh is None:
            thresh = self.thresh
        result = self.learn.model.predict_class_batch(
            text, self._device, self.is_multilabel_problem, thresh
        )
        return result

    def predict(
        self,
        text_or_list,
        show_progress=True,
        thresh=None,
        explain=False,
        explain_index=None,
        batch_size=64,
    ):
        """
        Predicts the class label(s) for the input text

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required String or List. text or a list of
                                texts for which we wish to find the class label(s).
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
                                Applicable only when a list of text is passed
        ---------------------   -------------------------------------------
        thresh                  Optional Float. The threshold value set to get
                                the class label(s). Applicable only for multi-label
                                classification task. Default is the value set
                                during the model creation time, otherwise the value
                                of 0.25 is set.
        ---------------------   -------------------------------------------
        explain                 Optional Bool. If set to True it shall generate SHAP
                                based explanation. Kindly visit:-
                                https://shap.readthedocs.io/en/latest/
        ---------------------   -------------------------------------------
        explain_index           Optional List. Index of the rows for which explanation
                                is required.  If the value is None, it will generate
                                an explanation for every row.
        ---------------------   -------------------------------------------
        batch_size              Optional integer.
                                Number of inputs to be processed at once.
                                Try reducing the batch size in case of out of
                                memory errors.
                                Default value : 64
        =====================   ===========================================

        :return: * In case of single label classification problem, a tuple containing the text, its predicted class label and the confidence score.

                 * In case of multi label classification problem, a tuple containing the text, its predicted class labels, a list containing 1's for the predicted labels, 0's otherwise and list containing a score for each label
        """
        if explain:
            try:
                import shap
            except:
                warnings.warn(
                    "SHAP is not installed. Model explainablity will not be available"
                )
                explain = False
                explain_index = None

        if self.is_multilabel_problem is False and thresh is not None:
            self.logger.error(
                "Passing a threshold value for non multi-label classification task "
                "will not have any affect on the predicting the class label"
            )

        sliced_text_list = []

        if isinstance(text_or_list, (list, tuple, np.ndarray)):
            preds = []
            if len(text_or_list) < batch_size:
                batch_size = len(text_or_list)

            remaining_len = len(text_or_list)
            if len(text_or_list) % batch_size == 0:
                iter_val = len(text_or_list) // batch_size
            else:
                iter_val = (len(text_or_list) // batch_size) + 1
            lower_range = 0
            upper_range = batch_size
            if show_progress:
                for ind in progress_bar(range(iter_val)):
                    prediction = self._predict_batch(
                        text_or_list[lower_range:upper_range], thresh
                    )
                    preds.extend(prediction)
                    remaining_len -= batch_size
                    lower_range = upper_range
                    if remaining_len >= batch_size:
                        upper_range = lower_range + batch_size
                    else:
                        upper_range = upper_range + remaining_len
            else:
                for ind in range(iter_val):
                    prediction = self._predict_batch(
                        text_or_list[lower_range:upper_range], thresh
                    )
                    preds.extend(prediction)
                    remaining_len -= batch_size
                    lower_range = upper_range
                    if remaining_len >= batch_size:
                        upper_range = lower_range + batch_size
                    else:
                        upper_range = upper_range + remaining_len
            result = [(text, *pred) for text, pred in zip(text_or_list, preds)]
            if explain:
                if isinstance(explain_index, int):
                    sliced_text_list = text_or_list[explain_index : explain_index + 1]
                elif isinstance(explain_index, (list, tuple, np.ndarray)):
                    for i in explain_index:
                        if i < len(text_or_list):
                            sliced_text_list.append(text_or_list[i])
                else:
                    sliced_text_list = copy.deepcopy(text_or_list)
                    warnings.warn(
                        "No Index is supplied. Going ahead with all the inputs"
                    )
        else:
            preds = self._predict(text_or_list, thresh)
            result = (text_or_list, *preds)
            if explain:
                sliced_text_list = [text_or_list]

        if explain:
            self._emodel, self._emask = self._wrapped_model_for_explnation()
            self._explain(sliced_text_list)

        return result

    def _save_df_to_html(self, path):
        if getattr(self._data, "_is_empty", False):
            if self._data.emd_path:
                copy_metrics(self._data.emd_path, path, model_characteristics_folder)
            return
        validation_dataframe = self._data._valid_df.sample(n=5)
        if self.is_multilabel_problem:
            predictions = [
                x[1]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist(),
                    show_progress=False,
                )
            ]
            labels = [
                ";".join(
                    [
                        column
                        for column in self._data._label_cols
                        if int(getattr(item, column))
                    ]
                )
                for idx, item in validation_dataframe.iterrows()
            ]

        else:
            predictions = [
                x[1]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist(),
                    show_progress=False,
                )
            ]
            labels = [x[0] for x in validation_dataframe[self._data._label_cols].values]

        new_df = pd.DataFrame(
            validation_dataframe[self._data._text_cols].values, columns=["source"]
        )
        new_df["target"] = labels
        new_df["predictions"] = predictions

        df_str = new_df.to_html(index=False, justify="left").replace(">\n", ">")

        msg = "<p><b>Sample Results</b></p>"

        text = f"\n\n{msg}\n\n{df_str}"

        save_data_in_model_metrics_html(text, path, model_characteristics_folder)

    def metrics_per_label(self):
        """
        :return: precision, recall and f1 score for each label in the classification model.
        """
        try:
            self._check_requisites()
        except Exception as e:
            metrics_per_label = self._data.emd.get("MetricsPerLabel")
            if metrics_per_label:
                metrics_per_label = json.loads(metrics_per_label)
                return self._create_dataframe_from_dict(metrics_per_label)
            else:
                self.logger.error("Metric not found in the loaded model")
        else:
            validation_dataframe = self._data._valid_df
            if self.is_multilabel_problem:
                predictions = [
                    x[2]
                    for x in self.predict(
                        validation_dataframe[self._data._text_cols].tolist()
                    )
                ]
                labels = [
                    [int(getattr(item, column)) for column in self._data._label_cols]
                    for idx, item in validation_dataframe.iterrows()
                ]
                target_names = self._data._label_cols
                output_dict = classification_report(
                    labels,
                    predictions,
                    target_names=target_names,
                    zero_division=1,
                    output_dict=True,
                )
            else:
                predictions = [
                    x[1]
                    for x in self.predict(
                        validation_dataframe[self._data._text_cols].tolist()
                    )
                ]
                labels = [
                    x[0] for x in validation_dataframe[self._data._label_cols].values
                ]
                target_names = self.learn.model._config.label2id.keys()
                if len(target_names) != len(set(labels)):
                    warnings.warn(
                        f'Validation dataset classes {list(set(labels))} does not match the training dataset \
classes {list(target_names)}, you could use "stratify=True" with prepare_textdata or try increasing the minority class \
samples. Metrics are only being calculated for classes present in the validation dataset.'
                    )
                    target_names = set(labels)
                output_dict = classification_report(
                    labels,
                    predictions,
                    target_names=target_names,
                    zero_division=1,
                    output_dict=True,
                )

            return self._create_dataframe_from_dict(output_dict)

    @staticmethod
    def _create_dataframe_from_dict(out_dict):
        out_dict.pop("accuracy", None)
        out_dict.pop("micro avg", None)
        out_dict.pop("macro avg", None)
        out_dict.pop("samples avg", None)
        out_dict.pop("weighted avg", None)
        df = pd.DataFrame(out_dict)
        # df.drop("support", inplace=True)
        dataframe = df.T.round(4)
        column_mappings = {
            "precision": "Precision_score",
            "recall": "Recall_score",
            "f1-score": "F1_score",
            "support": "Support",
        }
        dataframe.rename(columns=column_mappings, inplace=True)
        return dataframe

    def get_misclassified_records(self):
        """
        :return: get misclassified records for this classification model.
        """
        self._check_requisites()
        validation_dataframe = self._data._valid_df
        misclassified_records, text_col, label_col = (
            [],
            self._data._text_cols,
            self._data._label_cols,
        )
        if self.is_multilabel_problem:
            predictions = [
                x[1]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist()
                )
            ]
            labels = [
                ";".join(
                    [
                        column
                        for column in self._data._label_cols
                        if int(getattr(item, column))
                    ]
                )
                for idx, item in validation_dataframe.iterrows()
            ]
            for index, row in validation_dataframe.iterrows():
                label, prediction = labels[index], predictions[index]
                if label != prediction:
                    misclassified_records.append((row[text_col], label, prediction))
        else:
            predictions = [
                x[1]
                for x in self.predict(
                    validation_dataframe[self._data._text_cols].tolist()
                )
            ]
            label_col = (
                label_col if isinstance(label_col, (str, bytes)) else label_col[0]
            )
            for index, row in validation_dataframe.iterrows():
                label, prediction = row[label_col], predictions[index]
                if label != prediction:
                    misclassified_records.append((row[text_col], label, prediction))

        return pd.DataFrame(
            misclassified_records, columns=[text_col, "Target", "Prediction"]
        )

    def _explain(self, text_or_list, custom_tok=True):
        # """
        # To Generate explanation for a single or a batch of input strings.
        #
        # This function is a wrapper around SHAP explainer function for the Language model.
        # It relies on two underlying methods.
        # 1. It will wrap the logits of the model
        # 2. It will produce single class as an output.
        # EntityRecognizer
        #
        # =====================   ===========================================
        # **Parameter**            **Description**
        # ---------------------   -------------------------------------------
        # text_or_list            Required String or List. text or a list of
        #                         texts for which we wish to find the class label(s).
        #
        # custom_tok              Setting this argument to True will return the explanation
        #                         based on the word boundary token.
        # =====================   ===========================================
        # :return: None
        #
        # """
        has_shap = True
        try:
            import shap
        except:
            has_shap = False
            warnings.warn(
                "SHAP is not installed. Model explainablity will not be available"
            )
        if has_shap:
            if isinstance(text_or_list, str):
                text_or_list = [text_or_list]
            elif not isinstance(text_or_list, list):
                raise Exception(f" This module takes string or list as an input")
            # Build custom masker
            masker = None
            if custom_tok:
                masker = shap.maskers.Text(custom_tokenizer)
            # create labels
            labels = sorted(
                self.learn.model._config.label2id,
                key=self.learn.model._config.label2id.get,
            )
            explainer = shap.Explainer(self._logit_wrapper, masker, output_names=labels)
            text_or_list = [
                text if len(text.split(" ")) > 1 else text + "  "
                for text in text_or_list
            ]
            self.shap_values = explainer(text_or_list)
            shap.plots.text(self.shap_values)

    def _wrapped_model_for_explnation(self):
        # """
        # It will return the wrapped transformer for the classification task. It will return the
        # transformer from learner as well as the tokenizer.
        # """
        return self.learn.model._transformer, self.learn.model._tokenizer

    def _logit_wrapper(self, input_sent):
        input_sent = list(input_sent)
        # This code is modified to accomodate the masking structure and making the implementation verbose.
        encoded_dict = self._emask(
            input_sent,
            max_length=self._seq_len,
            pad_to_max_length=True,
            return_attention_mask=True,
            return_tensors="pt",
        )
        logits = self._emodel(
            encoded_dict["input_ids"].cuda(), encoded_dict["attention_mask"].cuda()
        )[0]
        results = torch.softmax(logits, dim=1).detach().cpu().numpy()
        return results



import json
import uuid
from urllib.parse import urlencode, urlparse, urlunparse, parse_qs, ParseResult
import xml.etree.cElementTree as ET
from io import BytesIO, StringIO

from arcgis.gis import GIS
from arcgis import env as _env
from arcgis._impl.common._mixins import PropertyMap

from ._base import BaseOGC


###########################################################################
class WMTSLayer(BaseOGC):
    """
    Represents a Web Map Tile Service, which is an OGC web service endpoint.


    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The web address of the endpoint.
    ---------------     --------------------------------------------------------------------
    version             Optional String. The version number of the WMTS service.  The default is `1.0.0`
    ---------------     --------------------------------------------------------------------
    gis                 Optional :class:`~arcgis.gis.GIS` . The GIS used to reference the service by. The arcgis.env.active_gis is used if not specified.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================



    """

    _gis = None
    _con = None
    _url = None
    _reader = None
    _cap_reader = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version="1.0.0", gis=None, **kwargs):
        super(WMTSLayer, self)
        if gis:
            gis = gis
        elif gis is None and _env.active_gis:
            gis = _env.active_gis
        else:
            gis = GIS()
        assert isinstance(gis, GIS)
        self._id = kwargs.pop("id", uuid.uuid4().hex)
        self._version = version
        self._con = gis._con
        self._title = kwargs.pop("title", "WMTS Layer")
        self._gis = gis
        if url[-1] == "/":
            url = url[:-1]
        self._url = url
        self._add_token = str(self._con._auth).lower() == "builtin"
        self._min_scale, self._max_scale = kwargs.pop("scale", (0, 0))
        self._opacity = kwargs.pop("opacity", 1)
        self._type = "WebTiledLayer"

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        Returns the properties of the Layer.

        :return: PropertyMap
        """
        if self._properties is None:
            from arcgis._impl.common._mixins import PropertyMap

            if self._add_token:
                url = self._capabilities_url(
                    service_url=self._url, vendor_kwargs={"token": self._con.token}
                )
            else:
                url = self._capabilities_url(service_url=self._url)
            text = self._con.get(url, {}, try_json=False, add_token=False)
            if text.find("Invalid Token") > -1 or text.find("Get Token") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<html>") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<?xml version=") > -1:
                pass
            else:
                raise Exception("Could not connect to the WebMap Tile Service")
            sss = BytesIO()
            sss.write(text.encode())
            sss.seek(0)
            tree = ET.XML(text=sss.read())
            d = self._xml_to_dictionary(tree)
            self._properties = PropertyMap(d)
        return self._properties

    # ----------------------------------------------------------------------
    def _capabilities_url(self, service_url, vendor_kwargs=None):
        """Return a capabilities url"""
        pieces = urlparse(service_url)
        args = parse_qs(pieces.query)
        if "service" not in args:
            args["service"] = "WMTS"
        if "request" not in args:
            args["request"] = "GetCapabilities"
        if "version" not in args:
            args["version"] = self._version
        if vendor_kwargs:
            args.update(vendor_kwargs)
        query = urlencode(args, doseq=True)
        pieces = ParseResult(
            pieces.scheme,
            pieces.netloc,
            pieces.path,
            pieces.params,
            query,
            pieces.fragment,
        )
        return urlunparse(pieces)

    # ----------------------------------------------------------------------
    def _format_tags(self, tag):
        """attempts to format tags by stripping out the {text} from the keys"""
        import re

        regex = r".*\}(.*)"
        matches = re.search(regex, tag)
        if matches:
            return matches.groups()[0]
        return tag

    # ----------------------------------------------------------------------
    def _xml_to_dictionary(self, t):
        """converts the xml to a dictionary object (recursivly)"""
        import json
        from collections import defaultdict

        d = {self._format_tags(t.tag): {} if t.attrib else None}
        children = list(t)
        if children:
            dd = defaultdict(list)
            for dc in map(self._xml_to_dictionary, children):
                for k, v in dc.items():
                    dd[self._format_tags(k)].append(v)
            d = {
                self._format_tags(t.tag): {
                    self._format_tags(k): v[0] if len(v) == 1 else v
                    for k, v in dd.items()
                }
            }
        if t.attrib:
            d[self._format_tags(t.tag)].update(
                [("@" + self._format_tags(k), v) for k, v in t.attrib.items()]
            )
        if t.text:
            text = t.text.strip()
            if children or t.attrib:
                if text:
                    d[self._format_tags(t.tag)]["#text"] = text
            else:
                d[self._format_tags(t.tag)] = text
        removals = [
            "{http://www.opengis.net/wmts/1.0}",
            "{http://www.opengis.net/ows/1.1}",
            "{http://www.w3.org/1999/xlink}",
        ]
        d = json.dumps(d)
        for remove in removals:
            d = d.replace(remove, "")
        return json.loads(d)

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        """Represents the MapView's widget JSON format"""
        return {
            "id": self._id,
            "title": self._title or "WMTS Layer",
            "url": self._url,
            "version": self._version,
            "minScale": self.scale[0],
            "maxScale": self.scale[1],
            "opacity": self.opacity,
            "type": self._type,
        }

    # ----------------------------------------------------------------------
    @property
    def __text__(self):
        """creates the item's text properties"""

        layer = None
        tile_matrix = None

        if isinstance(self.properties.Capabilities.Contents.Layer, (list, tuple)):
            layer = self.properties.Capabilities.Contents.Layer[0]
            tile_matrix = self.properties.Capabilities.Contents.TileMatrixSet[0]
        elif isinstance(
            self.properties.Capabilities.Contents.Layer, (dict, PropertyMap)
        ):
            layer = self.properties.Capabilities.Contents.Layer
            tile_matrix = self.properties.Capabilities.Contents.TileMatrixSet
        else:
            raise ValueError("Could not parse the results properly.")

        url_template = (
            layer.ResourceURL["@template"]
            .replace("{TileMatrix}", "{level}")
            .replace("{Style}", layer.Style.Identifier)
            .replace("{TileRow}", "{row}")
            .replace("{TileCol}", "{col}")
            .replace("{TileMatrixSet}", tile_matrix.Identifier)
        )
        fullExtent = [
            float(coord) for coord in layer.BoundingBox.LowerCorner.strip().split(" ")
        ] + [float(coord) for coord in layer.BoundingBox.UpperCorner.strip().split(" ")]
        lods = []
        WMTS_DPI = 90.71428571428571
        for l in tile_matrix.TileMatrix:
            lods.append(
                {
                    "level": int(l.Identifier),
                    "levelValue": l.Identifier,
                    "resolution": float(l.ScaleDenominator) * 0.00028,
                    "scale": float(l.ScaleDenominator) * WMTS_DPI / 96,
                }
            )
        return {
            "templateUrl": url_template,
            "copyright": "",
            "fullExtent": {
                "xmin": fullExtent[0],
                "ymin": fullExtent[1],
                "xmax": fullExtent[2],
                "ymax": fullExtent[3],
                "spatialReference": {
                    "wkid": int(layer.BoundingBox["@crs"].split(":")[-1])
                },
            },
            "tileInfo": {
                "rows": 256,
                "cols": 256,
                "dpi": 96,
                "origin": {
                    "x": (fullExtent[2] + fullExtent[0]) / 2,
                    "y": (fullExtent[3] + fullExtent[1]) / 2,
                    "spatialReference": {
                        "wkid": int(layer.BoundingBox["@crs"].split(":")[-1])
                    },
                },
                "spatialReference": {
                    "wkid": int(layer.BoundingBox["@crs"].split(":")[-1])
                },
                "lods": lods,
            },
            "wmtsInfo": {
                "url": self._url,
                "layerIdentifier": layer.Title,
                "tileMatrixSet": [tile_matrix.Identifier],
            },
        }

    @property
    def _operational_layer_json(self):
        """Represents the WebMap's JSON format"""
        return self.__text__
